function P = pinvarianti(C)
%PINVARIANTI   Calcolo dei P-invarianti di una rete di Petri
%
%   function P = pinvarianti(C)
%   Calcola i P-invarianti di una rete di Petri avente C come matrice di incidenza.
%   Restituisce una matrice di dimensioni 
%      <numero di P-invarianti> * <numero di posti della rete>
%   in cui ogni riga rappresenta un P-invariante.

%disp('Inizio Pinvarianti'),pause

cont_righe=0;  % conta le righe via via generate dall'algoritmo

trovato=0;  % flag: indica che si  trovata la colonna da cui parte l'algoritmo principale
strart_col=0;  % colonna di partenza ottenuta calcolando l'"expansion factor"
start_col2=0;  % colonna di partenza che minimizza il prodotto Pk*Nk (definiti sotto)
i=0;  % contatore per i prodotti parziali Pk*Nk

[numposti, numtrans] = size(C);
UO = sparse([C eye(numposti)]);  % creo la matrice UO 
P=[];  % matrice che conterr i p-invarianti

ann=0;  % contatore colonne annullate
col_ann=0;  % indica quale colonna  stata annullata 
c=0;  % contatore delle colonne da annullare
% conto le colonne da annullare
for k = 1:numtrans  %per ogni colonna di C
   pos=find(C(:,k)>0);  % indici degli elementi positivi della k-esima colonna di C
   Pk = length(pos);  % numero degli elementi positivi della k-esima colonna di C
   neg=find(C(:,k)<0);  % indici degli elementi negativi della k-esima colonna di C
   Nk = length(neg);  % numero degli elementi negativi della k-esima colonna di C
   if Pk * Nk == 1 & abs(UO(pos,k)) == abs(UO(neg,k))  % se Pk=Nk=1 e l'elemento positivo
      																 % e negativo sono uguali in modulo
      c=c+1;
   end
end

% prima ottimizzazione: viene annullata una colonna di UO che abbia un elemento
% positivo e un elemento negativo uguali in modulo, a meno che la colonna con queste 
% caratteristiche non sia una sola 
for k = 1:numtrans  %per ogni colonna di A
   pos=find(UO(:,k)>0);  % indici degli elementi positivi della k-esima colonna di A
   Pk = length(pos);  % numero degli elementi positivi della k-esima colonna di A
   neg=find(UO(:,k)<0);  % indici degli elementi negativi della k-esima colonna di A
   Nk = length(neg);  % numero degli elementi negativi della k-esima colonna di A
   % se Pk=Nk=1 e non si  ancora annullata la colonna di C e l'elemento 
   % positivo e il negativo sono uguali in modulo e la colonna da annullare non  una sola
	if Pk * Nk == 1 & abs(UO(pos,k)) == abs(UO(neg,k)) & ann < 1 & ann < 1 & c > 1 
    	UO(pos,k) = 0;  % annullo la
      UO(neg,k) = 0;  % k-esima colonna
      Pk=0;  % ora Pk  nullo
      Nk=0;  % e anche Nk
      ann=ann+1;  %  stata annullata una colonna
      col_ann=k;  % indica quale colonna  stata annullata
   end
end

% viene ora ricercata la colonna di partenza per la parte principale dell'algoritmo
for k = 1:numtrans  %per ogni colonna di A
   pos=find(C(:,k)>0);  % indici degli elementi positivi della k-esima colonna di C
   Pk = length(pos);  % numero degli elementi positivi della k-esima colonna di C
   neg=find(C(:,k)<0);  % indici degli elementi negativi della k-esima colonna di C
   Nk = length(neg);  % numero degli elementi negativi della k-esima colonna di C
   if Pk + Nk ~= 0 & Pk*Nk-(Pk+Nk) < 0 & trovato == 0  
      % se la colonna  non nulla, l'"expansion factor"  negativo e non ho ancora trovato
      % la colonna di partenza
      trovato=1;  % ho trovato la colonna di partenza
      start_col=k;  
   end
   if Pk + Nk ~= 0 & trovato == 0  % se la colonna  non nulla e
         				% non ha un "expansion factor" negativo
      prod_Pk_Nk=Pk*Nk;  % calcolo Pk*Nk
      if i == 0  % se  la prima colonna per cui calcolo Pk*Nk
         slack=prod_Pk_Nk;  % setto una variabile di confronto a Pk*Nk
         i=1;  % aggiorno il contatore
         start_col2=k;  % setto la colonna di partenza secondo questa logica
      end
      if prod_Pk_Nk < slack  % se trovo un prodotto Pk*Nk minore di quello
            					  % calcolato precedentemente
         slack=prod_Pk_Nk;  % aggiorno la variabile di confronto
         start_col2=k;  % e la colonna di partenza
      end   
   end
end   
if trovato == 0  % se non ho trovato una colonna che soddisfa il criterio
   				  % dell'"expansion factor" negativo
   start_col=start_col2;  % la colonna di partenza  quella che minimizza Pk*Nk 
end

Uk1=UO;  % poni la matrice Uk1 = UO

act_ann=zeros(1,numtrans);  % maschera della colonne attivamente annullate

[numpostiUk1, numcolUk1]=size(Uk1); % serve solo "numpostiUk1" per l'algoritmo principale;
												% mostra a video le due dimensioni
k=start_col;  % mostra la colonna di partenza dell'algoritmo principale

act_ann(k)=1;  % metti a 1 il valore della maschera corrispondente alla colonna k

Uk=sparse([]);  % inizializza la matrice Uk
Uneg=sparse([]);  % inizializza la matrice Uneg
Upos=sparse([]);  % inizializza la matrice Upos
while k ~= 0  % inizio dell'algoritmo principale
   for j = 1:numpostiUk1  % per ogni riga di Uk1
      if Uk1(j, k) == 0  % se l'elemento all'incrocio fra riga j e colonna k  pari a 0
         Uk=[Uk; Uk1(j, :)];  % aggiungi la riga con esso ad Uk  
      elseif Uk1(j, k) < 0  % se l'elemento all'incrocio fra riga j e colonna k  < 0
			Uneg=[Uneg; Uk1(j, :)];  % aggiungi la riga con esso ad Uneg
      else  % se l'elemento all'incrocio fra riga j e colonna k  > 0
		   Upos=[Upos; Uk1(j, :)];  % aggiungi la riga con esso ad Upos
      end
   end   
   
   % se Uneg e/o Upos sono vuote, metti a 0 il valore della maschera corrispondente 
   % alla colonna k 
   if isequal(Uneg,[])
      act_ann(k)=0;
   end
   if isequal(Upos,[])
      act_ann(k)=0;
   end   
   
   [x1, y1]=size(Uneg);  % serve solo x1: numero delle righe di Uneg
   [x2, y2]=size(Upos);  % serve solo x2: numero delle righe di Upos
   [xk, yk]=size(Uk);  % serve solo xk: numero delle righe di Uk  
   if x1 ~= 0 & x2 ~= 0  % se le matrici non hanno dimensioni nulle
      row1=1;  % indice prima riga di Uneg
      while row1 <= x1
         row2=1;  % indice prima riga di Upos
         while row2 <= x2
            % calcola il supporto del nuovo P-i Y come unione dei supporti dei P-I
            % corrispondenti alle righe row1 e row2 rispettivamente di Uneg e di Upos
            suppY=spones(Uneg(row1, numtrans+1:numposti+numtrans))|spones(Upos(row2, numtrans+1:numposti+numtrans));
            IsuppY=find(suppY>0);  % indici elementi diversi da zero del supporto di Y
            card_suppY=length(find(IsuppY>0));  % calcola la cardinalit del supporto di Y
            supp_rowC=zeros(1, numtrans);  % vettore che conterr l'unione dei supporti delle
            									    % righe di C indicizzate dal supporto di Y
            % calcola l'unione dei supporti delle righe di C indicizzate dal supporto di Y                            
            for f = IsuppY  % per ogni valore del supporto di Y
               supp_rowC=spones(C(f, :))|supp_rowC; % calcola l'unione dei corrispondenti
               												 % supporti delle righe di C
            end
            
            % calcola il rank_upperbound
            rank_upperbound=length(find(act_ann.*supp_rowC>0));
            % ottimizzazione: esegue il "fast test of minimality"
				if card_suppY <= rank_upperbound + 1
            	v=1;  % contatore   
               includes=0;  % flag usato nel "support comparison"
               % esecuzione del "support comparison": confronto il supporto del P-I trovato
               % con il supporto dei P-I presenti in Uk, Uneg o Upos
               % confronto con i P-I di Uneg:
               while v <= x1 & includes == 0
                  if v ~= row1
                     % esegui il "support comparison" fra
                     % il nuovo P-I e quelli trovati in precedenza
                     supp_comp=suppY|Uneg(v,numtrans+1:numposti+numtrans);   
                  	if supp_comp == suppY  % se il nuovo P-I contiene il supporto di altri
                  								  % P-I trovati prima                              
                     	v=x1+1;  % termina il "support comparison"
                     	includes=1;  % il nuovo P-I ne contiene un altro 
                  	else
                     	v=v+1;  % altrimenti continua
                  	end
                  else   
                  	v=v+1;   
                  end
               end   
               % confronto con i P-i di Upos:
               v=1;
               while v <= x2 & includes == 0 
                  if v ~= row2
                     % esegui il "support comparison" fra 
                  	% il nuovo P-I e quelli trovati in precedenza
                     supp_comp=suppY|Upos(v,numtrans+1:numposti+numtrans);                    	
                     if supp_comp == suppY  % se il nuovo P-I contiene il supporto di altri
                  								  % P-I trovati prima                              
                     	v=x2+1;  % termina il "support comparison"
                     	includes=1;  % il nuovo P-I ne contiene un altro
                  	else
                     	v=v+1;  % altrimenti continua
                  	end
                  else
                     v=v+1;
                  end
               end
               % confronto con i P-i di Uk:
               v=1;
               while v <= xk & includes == 0
                  % esegui il "support comparison" fra 
                 	% il nuovo P-I e quelli trovati in precedenza
                  supp_comp=suppY|Uk(v,numtrans+1:numposti+numtrans);  
                  if supp_comp == suppY  % se il nuovo P-I contiene il supporto di altri
                  							  % P-I trovati prima                              
                     v=xk+1;  % termina il "support comparison"
                     includes=1;  % il nuovo P-I ne contiene un altro
                  else
                     v=v+1;  % altrimenti continua
                  end
               end
               if includes ~= 1  % se il P-I non contiene il supporto di altri P-I
                  alpha=abs(Upos(row2, k));  % calcola il coefficiente alpha
                  beta=abs(Uneg(row1, k));  % calcola il coefficiente beta
                  new_row=alpha*Uneg(row1, :)+beta*Upos(row2, :);  % calcola la nuova riga
                  																 % da aggiungere a Uk
                                                                   
                  cont_righe=cont_righe+1;  % calcola le righe generate fino ad ora
                  
                  % calcola il MCD della nuova riga                                                 
                  supp_new_row=find(new_row);  
                  MCD=new_row(supp_new_row(1));
                  for l = 2:length(supp_new_row)
                     MCD=gcd(MCD, new_row(supp_new_row(l)));
                  end
                  Uk=[Uk; new_row/MCD];  % aggiungi a Uk la nuova riga divisa per il MCD
               end
            end
            row2=row2+1;  % passa alla successiva riga di Upos
         end
      row1=row1+1;  % passa alla successiva riga di Uneg
      end
   end   
   Uneg=sparse([]);  % annulla la matrice Uneg
   Upos=sparse([]);  % annulla la matrice Upos
   
   
   trovato=0;  % % flag: indica che si  trovata la colonna da cui parte l'algoritmo principale
   strart_col=0;  % colonna di partenza ottenuta calcolando l'"expansion factor"
	start_col2=0;  % colonna di partenza che minimizza il prodotto Pk*Nk (definiti sotto)
	i=0;  % contatore per i prodotti parziali Pk*Nk
   [numuk,num2uk]=size(Uk);
   if numuk == 0
      start_col=0;
   else
   % viene ora ricercata la colonna di partenza per la parte principale dell'algoritmo
		for k = 1:numtrans  %per ogni colonna di A
      	pos=find(Uk(:,k)>0);  % indici degli elementi positivi della k-esima colonna di C
   		Pk = length(pos);  % numero degli elementi positivi della k-esima colonna di C
   		neg=find(Uk(:,k)<0);  % indici degli elementi negativi della k-esima colonna di C
         Nk = length(neg);  % numero degli elementi negativi della k-esima colonna di C
      	if Pk + Nk ~= 0 & Pk*Nk-(Pk+Nk) < 0 & trovato == 0  
      		% se la colonna  non nulla, l'"expansion factor"  negativo e non ho ancora trovato
      		% la colonna di partenza
      		trovato=1;  % ho trovato la colonna di partenza
      		start_col=k;  
   		end
   		if Pk + Nk ~= 0 & trovato == 0  % se la colonna  non nulla e
      		   								  % non ha un "expansion factor" negativo
      		prod_Pk_Nk=Pk*Nk;  % calcolo Pk*Nk
      		if i == 0  % se  la prima colonna per cui calcolo Pk*Nk
         		slack=prod_Pk_Nk;  % setto una variabile di confronto a Pk*Nk
         		i=1;  % aggiorno il contatore
         		start_col2=k;  % setto la colonna di partenza secondo questa logica
      		end
      		if prod_Pk_Nk < slack  % se trovo un prodotto Pk*Nk minore di quello
         		   					  % calcolato precedentemente
         		slack=prod_Pk_Nk;  % aggiorno la variabile di confronto
         		start_col2=k;  % e la colonna di partenza
      		end   
   		end
		end   
		if trovato == 0  % se non ho trovato una colonna che soddisfa il criterio
   						  % dell'"expansion factor" negativo
   		start_col=start_col2;  % la colonna di partenza  quella che minimizza Pk*Nk 
		end
   end   
   k=start_col;  % mostra la colonna di partenza dell'algortitmo principale
   
   if k ~= 0  % se k  diverso da 0, poni la posizione della maschera corrispondente a k pari a 1
      act_ann(k)=1;
   end
   
	Uk1=Uk;  % prepara la matrice di partenza Uk1 per l'algoritmo principale
	[numpostiUk1, numtransUk1]=size(Uk1);  % serve solo "numpostiUk1" per l'algoritmo principale
   Uk=sparse([]);  % setta Uk come matrice nulla
end

% calcola la matrice dei P-I, P
for n=1:numpostiUk1
	if Uk1(n,1:numtrans) == zeros(1,numtrans);
   	P=full([Uk1(n,numtrans+1:numposti+numtrans); P]);  % metti i P-I trovati in P
	end
end

