function S = sifoni(C)
%SIFONI   Calcolo dei sifoni minimi di una rete di Petri
%
%   function S = sifoni(C)
%   Calcola i sifoni minimi di una rete di Petri avente C come matrice di incidenza.
%   Non riporta come soluzioni i supporti di P-invarianti della rete.
%   Restituisce una matrice di dimensioni 
%      <numero di sifoni> * <numero di posti della rete>
%   in cui ogni riga rappresenta un sifone.

PI = pinvarianti(C);
S = [];

NP = size(C,1);
for indp=1:NP
  % disp(indp)
  CS = zeros(1,NP);
  CS(indp) = 1;

  while ~isempty(CS)
    CSnew = [];
    for i=1:size(CS,1)
      [CS2,S2] = postisif(C,PI,CS(i,:),indp);
      CSnew = [CSnew; CS2];
      S = prune([S;S2],PI);
    end
    CS = CSnew;
  end;

end;
